
#include <time.h>
#include <math.h>
#include <stdlib.h>

/* Definition of HildonStatusBarPluginFn_st */
#include <hildon-status-bar-lib/hildon-status-bar-item.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixdata.h>

#include "panelclock-face.h"

#define COMPOSITE(a,b,v) ((guchar) (((guint16) (a) * (v) +\
                                     (guint16) (b) * (255 - (v)))\
                                    >> 8))

#define CLOCK_CENTER_X 20
#define CLOCK_CENTER_Y 19
#define CLOCK_RADIUS_M 11
#define CLOCK_RADIUS_H 7

static void * panelclock_initialize   (HildonStatusBarItem  *item,
                                       GtkWidget           **button);
static void   panelclock_destroy      (void                 *data);
static void   panelclock_update       (void                 *data,
                                       gint                  value1,
                                       gint                  value2,
                                       const gchar          *str);
static int    panelclock_get_priority (void                 *data);

struct _ClockPlugin
{
  HildonStatusBarItem *item;
  GtkWidget           *image;
  GdkPixdata          *face_pixdata;
  guint                timeout_id;
  /* Add here any data for your plugin. */
};

typedef struct _ClockPlugin ClockPlugin;


void
panelclock_entry (HildonStatusBarPluginFn_st *fn)
{
  if (fn == NULL) {
    return;
  }
  fn->initialize = panelclock_initialize;
  fn->destroy = panelclock_destroy;
  fn->update = panelclock_update;
  fn->get_priority = panelclock_get_priority;
}


static void
gdk_pixbuf_put_pixel (GdkPixbuf *pixbuf,
                      int x, int y,
                      guchar red, guchar green, guchar blue, guchar alpha)
{
  int width, height, rowstride, n_channels;
  guchar *pixels, *p;

  n_channels = gdk_pixbuf_get_n_channels (pixbuf);

  g_assert (gdk_pixbuf_get_colorspace (pixbuf) == GDK_COLORSPACE_RGB);
  g_assert (gdk_pixbuf_get_bits_per_sample (pixbuf) == 8);
  g_assert (gdk_pixbuf_get_has_alpha (pixbuf));
  g_assert (n_channels == 4);

  width = gdk_pixbuf_get_width (pixbuf);
  height = gdk_pixbuf_get_height (pixbuf);

  g_assert (x >= 0 && x < width);
  g_assert (y >= 0 && y < height);

  rowstride = gdk_pixbuf_get_rowstride (pixbuf);
  pixels = gdk_pixbuf_get_pixels (pixbuf);

  p = pixels + y * rowstride + x * n_channels;
  p[0] = COMPOSITE (red, p[0], alpha);
  p[1] = COMPOSITE (green, p[1], alpha);
  p[2] = COMPOSITE (blue, p[2], alpha);
}


static void
gdk_pixbuf_draw_line_aa (GdkPixbuf *pixbuf,
                         gint x0, gint y0, gint x1, gint y1,
                         guchar red, guchar green, guchar blue)
{
  gint dx, dy;
  gint x, y;
  gint slant;

  dx = x1 - x0;
  dy = y1 - y0;

  if (dx == 0 && dy == 0)
    {
      gdk_pixbuf_put_pixel (pixbuf, x0, y0, red, green, blue, 255);
      return;
    }

  if (abs (dx) > abs (dy))
    {
      if (x1 < x0)
        {
          x = x0; x0 = x1; x1 = x;
          y = y0; y0 = y1; y1 = y;
          dx = x1 - x0;
          dy = y1 - y0;
        }

      slant = (dy << 8) / dx;

      y = y0 << 8;

      for (x = x0; x <= x1; x++)
        {
          gdk_pixbuf_put_pixel (pixbuf,
                                x, y >> 8,
                                red, green, blue,
                                255 - (y % 256));
          gdk_pixbuf_put_pixel (pixbuf,
                                x, (y >> 8) + 1,
                                red, green, blue,
                                y % 256);
          y += slant;
        }
    }
  else
    {
      if (y1 < y0)
        {
          x = x0; x0 = x1; x1 = x;
          y = y0; y0 = y1; y1 = y;
          dx = x1 - x0;
          dy = y1 - y0;
        }

      slant = (dx << 8) / dy;

      x = x0 << 8;

      for (y = y0; y <= y1; y++)
        {
          gdk_pixbuf_put_pixel (pixbuf,
                                x >> 8, y,
                                red, green, blue,
                                255 - (x % 256));
          gdk_pixbuf_put_pixel (pixbuf,
                                (x >> 8) + 1, y,
                                red, green, blue,
                                x % 256);
          x += slant;
        }
    }
}


static GdkPixbuf *
panelclock_create_face (ClockPlugin *panelclock)
{
  GdkPixbuf *pixbuf = NULL;
  GError    *error  = NULL;
  gdouble    angle;
  time_t     t;
  struct tm  time_info;

  pixbuf = gdk_pixbuf_from_pixdata (panelclock->face_pixdata,
                                    TRUE, &error);
  if (error) g_printerr ("%s\n", error->message);

  t = time (NULL);
  localtime_r (&t, &time_info);

  angle = (gdouble) time_info.tm_min / 30.0 * G_PI;
  gdk_pixbuf_draw_line_aa (pixbuf,
                           CLOCK_CENTER_X, CLOCK_CENTER_Y,
                           CLOCK_CENTER_X + sin (angle) * CLOCK_RADIUS_M + 0.5,
                           CLOCK_CENTER_Y - cos (angle) * CLOCK_RADIUS_M + 0.5,
                           80, 80, 80);

  angle = (gdouble) (time_info.tm_hour * 60 + time_info.tm_min) / 360.0 * G_PI;
  gdk_pixbuf_draw_line_aa (pixbuf,
                           CLOCK_CENTER_X, CLOCK_CENTER_Y,
                           CLOCK_CENTER_X + sin (angle) * CLOCK_RADIUS_H + 0.5,
                           CLOCK_CENTER_Y - cos (angle) * CLOCK_RADIUS_H + 0.5,
                           0, 0, 0);

  return pixbuf;
}


static gboolean
panelclock_update_time (gpointer data)
{
  ClockPlugin *panelclock = (ClockPlugin *) data;
  GdkPixbuf *pixbuf;

  pixbuf = panelclock_create_face (panelclock);
  gtk_image_set_from_pixbuf (GTK_IMAGE (panelclock->image),
                             pixbuf);
  g_object_unref (pixbuf);

  return TRUE;
}


static void *
panelclock_initialize (HildonStatusBarItem  *item,
                       GtkWidget           **widget)
{
  GtkWidget   *image = NULL;
  GdkPixbuf   *pixbuf;
  GError      *error = NULL;
  ClockPlugin *panelclock;

  panelclock = g_new0 (ClockPlugin, 1);

  panelclock->item = item;

  panelclock->face_pixdata = g_new (GdkPixdata, 1);
  gdk_pixdata_deserialize (panelclock->face_pixdata,
                           1780, clock_face, &error);
  if (error) g_printerr ("%s\n", error->message);

  pixbuf = panelclock_create_face (panelclock);
  image = gtk_image_new_from_pixbuf (pixbuf); /* Should be 40x40 */
  g_object_unref (G_OBJECT (pixbuf));

  *widget = panelclock->image = image;

  panelclock->timeout_id = g_timeout_add (15 * 1000,
                                          panelclock_update_time,
                                          panelclock);

  gtk_widget_show_all (panelclock->image);

  return (void *) panelclock;
}


static void
panelclock_update (void        *data,
                   gint         value1,
                   gint         value2,
                   const gchar *str)
{
  if (!data) {
    return;
  }

  ClockPlugin *panelclock = (ClockPlugin *) data;

  if (value1 == 0) {
    /* Destroy the applet, we get unloaded by this. Note, destroying the item,
       not the button. */
    gtk_widget_destroy (GTK_WIDGET (panelclock->item));
  }
  else {
    /* Reveal the plugin. */
    gtk_widget_show_all (panelclock->image);
  }

  return;
}


static void
panelclock_destroy (void *data)
{
  ClockPlugin *panelclock = (ClockPlugin *) data;

  if (!data) {
    return;
  }

  g_source_remove (panelclock->timeout_id);
  g_free (data);
}


static gint
panelclock_get_priority (void *data)
{
  return 42;
}

